﻿Option Strict On
Option Explicit On

Imports System.Runtime.InteropServices

Public Class FaceDetection
    Implements IDisposable

#Region "API"
    <DllImport("Combase.dll", EntryPoint:="WindowsCreateString")>
    <PreserveSig> Private Shared Function WindowsCreateString(<[In], MarshalAs(UnmanagedType.LPWStr)> sourceString As String,
                                                              <[In]> length As Integer,
                                                              <Out> ByRef hString As IntPtr) As Integer
    End Function

    <DllImport("Combase.dll", EntryPoint:="WindowsDeleteString")>
    <PreserveSig> Private Shared Function WindowsDeleteString(<[In]> hString As IntPtr) As Integer
    End Function

    <DllImport("Combase.dll", EntryPoint:="RoGetActivationFactory")>
    <PreserveSig> Private Shared Function RoGetActivationFactory(<[In]> activatableClassId As IntPtr,
                                                                 <[In], MarshalAs(UnmanagedType.LPStruct)> riid As Guid,
                                                                 <Out> ByRef factory As IntPtr) As Integer

    End Function

    <DllImport("Shlwapi.dll", EntryPoint:="SHCreateMemStream")>
    Private Shared Function SHCreateMemStream(<[In]> pInit As IntPtr,
                                              <[In]> cbInit As UInteger) As IntPtr
    End Function

    <DllImport("ShCore.dll", EntryPoint:="CreateRandomAccessStreamOverStream")>
    <PreserveSig> Private Shared Function CreateRandomAccessStreamOverStream(<[In]> pStream As IntPtr,
                                                                             <[In]> options As BSOS_OPTIONS,
                                                                             <[In], MarshalAs(UnmanagedType.LPStruct)> riid As Guid,
                                                                             <Out> ByRef ppv As IntPtr) As Integer
    End Function

    <DllImport("GdiPlus.dll", EntryPoint:="GdipCreateBitmapFromHBITMAP")>
    Private Shared Function GdipCreateBitmapFromHBITMAP(<[In]> hbm As IntPtr,
                                                        <[In]> hpal As IntPtr,
                                                        <Out> ByRef pBitmap As IntPtr) As Integer
    End Function

    <DllImport("GdiPlus.dll", EntryPoint:="GdipDisposeImage")>
    Private Shared Function GdipDisposeImage(<[In]> pImage As IntPtr) As Integer
    End Function

    <DllImport("GdiPlus.dll", EntryPoint:="GdipSaveImageToStream")>
    Private Shared Function GdipSaveImageToStream(<[In]> pImage As IntPtr,
                                                  <[In]> pStream As IntPtr,
                                                  <[In]> ByRef ClsidEncoder As Guid,
                                                  <[In]> EncoderParams As IntPtr) As Integer
    End Function

    <DllImport("Gdi32.dll", EntryPoint:="DeleteObject")>
    Private Shared Function DeleteObject(<[In]> hObject As IntPtr) As <MarshalAs(UnmanagedType.Bool)> Boolean
    End Function
#End Region

#Region "Enum"
    Private Enum VTable_Interfaces As Integer

        'IReadOnlyList
        IReadOnlyList_GetAt = 6
        IReadOnlyList_GetSize = 7

        ' IBitmapDecoderStatics
        IBitmapDecoderStatics_CreateAsync = 14

        ' IBitmapFrame
        IBitmapFrame_GetBitmapPixelFormat = 8

        ' IBitmapFrameWithSoftwareBitmap
        IBitmapFrameWithSoftwareBitmap_GetSoftwareBitmapAsync = 6

        ' ISoftwareBitmapStatics
        ISoftwareBitmapStatics_Convert = 7

        'IFaceDetectorStatics
        IFaceDetectorStatics_CreateAsync = 6
        IFaceDetectorStatics_GetSupportedBitmapPixelFormats = 7

        ' IFaceDetector
        IFaceDetector_DetectFacesAsync = 6

        ' IDetectedFace
        IDetectedFace_GetFaceBox = 6

        ' IAsyncInfo
        IAsyncInfo_GetStatus = 7
        IAsyncInfo_Close = 10

        ' IAsyncInterface
        IAsyncInterface_GetResults = 8

        ' IClosable
        IClosable_Close = 6
    End Enum

    Private Enum FileAccessMode As Integer
        Read = 0
        ReadWrite = 1
    End Enum

    Private Enum BSOS_OPTIONS As Integer
        BSOS_DEFAULT = 0
        BSOS_PREFERDESTINATIONSTREAM = 1
    End Enum

    Private Enum AsyncStatus As Integer
        Started = 0
        Completed = 1
        Canceled = 2
        [Error] = 3
    End Enum

    Public Enum BitmapPixelFormat As Integer
        BitmapPixelFormat_Unknown = 0
        BitmapPixelFormat_Rgba16 = 12
        BitmapPixelFormat_Rgba8 = 30
        BitmapPixelFormat_Gray16 = 57
        BitmapPixelFormat_Gray8 = 62
        BitmapPixelFormat_Bgra8 = 87
        BitmapPixelFormat_Nv12 = 103
        BitmapPixelFormat_Yuy2 = 107
    End Enum

#End Region

#Region "Const"
    Private Const S_OK As Integer = 0
    Private Const GDIP_OK As Integer = 0

    Private Const WindowsGraphicsImagingBitmapDecoder As String = "Windows.Graphics.Imaging.BitmapDecoder"
    Private Const WindowsGraphicsImagingSoftwareBitmap As String = "Windows.Graphics.Imaging.SoftwareBitmap"
    Private Const WindowsMediaFaceAnalysisFaceDetector As String = "Windows.Media.FaceAnalysis.FaceDetector"

    Private Const IID_IClosable As String = "30d5a829-7fa4-4026-83bb-d75bae4ea99e"
    Private Const IID_IAsyncInfo As String = "00000036-0000-0000-c000-000000000046"
    Private Const IID_IBitmapFrame As String = "72a49a1c-8081-438d-91bc-94ecfc8185c6"
    Private Const IID_IRandomAccessStream As String = "905a0fe1-bc53-11df-8c49-001e4fc686da"
    Private Const IID_IFaceDetectorStatics As String = "bc042d67-9047-33f6-881b-6746c1b218b8"
    Private Const IID_IBitmapDecoderStatics As String = "438ccb26-bcef-4e95-bad6-23a822e58d01"
    Private Const IID_ISoftwareBitmapStatics As String = "df0385db-672f-4a9d-806e-c2442f343e86"
    Private Const IID_IBitmapFrameWithSoftwareBitmap As String = "fe287c9a-420c-4963-87ad-691436e08383"

    Private Const CLSID_GDIPPngEncoder As String = "557cf406-1a04-11d3-9a73-0000f81ef32e"
#End Region

#Region "Class"
    Public Sub New()
        If GetActivationFactory(WindowsGraphicsImagingBitmapDecoder,
                                IID_IBitmapDecoderStatics,
                                m_pIBitmapDecoderStatics) Then
            If GetActivationFactory(WindowsGraphicsImagingSoftwareBitmap,
                                    IID_ISoftwareBitmapStatics,
                                    m_pISoftwareBitmapStatics) Then
                Dim pIFaceDetectorStatics As IntPtr = IntPtr.Zero
                If GetActivationFactory(WindowsMediaFaceAnalysisFaceDetector,
                                        IID_IFaceDetectorStatics,
                                        pIFaceDetectorStatics) Then
                    If FaceDetectorStatics_CreateAsync(pIFaceDetectorStatics,
                                                       m_pIFaceDetector) Then
                        If WaitForAsync(m_pIFaceDetector) Then
                            m_bolIsFaceDetectorInitialized = True
                        End If
                    End If
                    GetSupportedBitmapPixelFormats(pIFaceDetectorStatics)
                    Release(pIFaceDetectorStatics)
                End If
            End If
        End If
    End Sub
#End Region

#Region "Variablen"
    Private m_bolDisposedValue As Boolean
    Private ReadOnly m_bolIsFaceDetectorInitialized As Boolean
    Private m_tDetectedFacesBounds As List(Of Rectangle)
    Private m_eBitmapPixelFormats As List(Of BitmapPixelFormat)
    Private m_pIFaceDetector As IntPtr
    Private m_pIBitmapDecoderStatics As IntPtr
    Private m_pISoftwareBitmapStatics As IntPtr
#End Region

#Region "Properties"
    Public ReadOnly Property GetIsFaceDetectorInitialized As Boolean
        Get
            Return m_bolIsFaceDetectorInitialized
        End Get
    End Property

    Public ReadOnly Property GetSupportedBitmapPixelFormatCount As Integer
        Get
            Return m_eBitmapPixelFormats.Count
        End Get
    End Property

    Public ReadOnly Property GetSupportedBitmapPixelFormatByIndex(Index As Integer) As BitmapPixelFormat
        Get
            If Index >= 0 AndAlso Index < Me.GetSupportedBitmapPixelFormatCount Then
                Return m_eBitmapPixelFormats(Index)
            Else
                Return New BitmapPixelFormat
            End If
        End Get
    End Property

    Public ReadOnly Property GetDetectedFacesCount As Integer
        Get
            Return m_tDetectedFacesBounds.Count
        End Get
    End Property

    Public ReadOnly Property GetDetectedFaceBoundByIndex(Index As Integer) As Rectangle
        Get
            If Index >= 0 AndAlso Index < Me.GetDetectedFacesCount Then
                Return m_tDetectedFacesBounds(Index)
            Else
                Return New Rectangle
            End If
        End Get
    End Property

    Public ReadOnly Property GetDetectedFacesBounds() As List(Of Rectangle)
        Get
            Return m_tDetectedFacesBounds
        End Get
    End Property
#End Region

#Region "Public Functions"
    Public Function GetFaceDetectionResultsFromBitmap(bitmap As Bitmap,
                                                      pixelFormat As BitmapPixelFormat) As Boolean
        Dim bolRet As Boolean = False
        If m_bolIsFaceDetectorInitialized AndAlso bitmap IsNot Nothing Then
            Dim hBitmap As IntPtr = bitmap.GetHbitmap
            If hBitmap <> IntPtr.Zero Then
                Dim pBitmap As IntPtr = IntPtr.Zero
                If GdipCreateBitmapFromHBITMAP(hBitmap, IntPtr.Zero, pBitmap) = GDIP_OK Then
                    Dim pIStream As IntPtr = SHCreateMemStream(IntPtr.Zero, 0)
                    If pIStream <> IntPtr.Zero Then
                        If GdipSaveImageToStream(pBitmap, pIStream,
                                                 New Guid(CLSID_GDIPPngEncoder),
                                                 IntPtr.Zero) = GDIP_OK Then
                            Dim pIRandomAccessStream As IntPtr = IntPtr.Zero
                            If CreateRandomAccessStreamOverStream(pIStream, BSOS_OPTIONS.BSOS_DEFAULT,
                                                                  New Guid(IID_IRandomAccessStream),
                                                                  pIRandomAccessStream) = S_OK Then
                                bolRet = GetFaceDetectionFromIRandomAccessStream(pIRandomAccessStream,
                                                                                 pixelFormat)
                                CloseAndRelease(pIRandomAccessStream)
                            End If
                        End If
                        Release(pIStream)
                    End If
                    GdipDisposeImage(pBitmap)
                End If
                DeleteObject(hBitmap)
            End If
        End If
        Return bolRet
    End Function
#End Region

#Region "Private Functions"
    Private Function GetFaceDetectionFromIRandomAccessStream(pIRandomAccessStream As IntPtr,
                                                             pixelFormat As BitmapPixelFormat) As Boolean
        Dim bolRet As Boolean = False
        Dim pISoftwareBitmap As IntPtr = IntPtr.Zero
        If GetSoftwareBitmapFromRandomAccessStream(pIRandomAccessStream,
                                                   pixelFormat,
                                                   pISoftwareBitmap) Then
            Dim pIReadOnlyList As IntPtr = IntPtr.Zero
            If FaceDetector_DetectFacesAsync(m_pIFaceDetector,
                                             pISoftwareBitmap,
                                             pIReadOnlyList) Then
                If WaitForAsync(pIReadOnlyList) Then
                    Dim intCount As Integer
                    If ReadOnlyList_GetSize(pIReadOnlyList,
                                            intCount) Then
                        If intCount > 0 Then
                            m_tDetectedFacesBounds = New List(Of Rectangle)
                            For intIndex As Integer = 0 To intCount - 1
                                Dim pIDetectedFace As IntPtr = IntPtr.Zero
                                If ReadOnlyList_GetAt(pIReadOnlyList,
                                                      intIndex,
                                                      pIDetectedFace) Then
                                    Dim tBitmapBounds As New Rectangle
                                    If DetectedFace_GetFaceBox(pIDetectedFace,
                                                               tBitmapBounds) Then
                                        m_tDetectedFacesBounds.Add(tBitmapBounds)
                                    End If
                                    Release(pIDetectedFace)
                                End If
                            Next
                            bolRet = True
                        End If
                    End If
                End If
                Release(pIReadOnlyList)
            End If
            CloseAndRelease(pISoftwareBitmap)
        End If
        Return bolRet
    End Function

    Private Function GetSoftwareBitmapFromRandomAccessStream(pIRandomAccessStream As IntPtr,
                                                             pixelFormat As BitmapPixelFormat,
                                                             ByRef pISoftwareBitmap As IntPtr) As Boolean
        Dim bolRet As Boolean = False
        If m_bolIsFaceDetectorInitialized Then
            Dim pIBitmapDecoder As IntPtr = IntPtr.Zero
            If BitmapDecoderStatics_CreateAsync(m_pIBitmapDecoderStatics,
                                                pIRandomAccessStream,
                                                pIBitmapDecoder) Then
                If WaitForAsync(pIBitmapDecoder) Then
                    Dim pIBitmapFrame As IntPtr = IntPtr.Zero
                    If QueryInterface(pIBitmapDecoder,
                                      IID_IBitmapFrame,
                                      pIBitmapFrame) Then
                        Dim eBitmapPixelFormat As New BitmapPixelFormat
                        If BitmapFrame_GetBitmapPixelFormat(pIBitmapFrame, eBitmapPixelFormat) Then
                            Dim bolIsBitmapPixelFormatSupported As Boolean = m_eBitmapPixelFormats.Contains(eBitmapPixelFormat)
                            Dim pIBitmapFrameWithSoftwareBitmap As IntPtr = IntPtr.Zero
                            If QueryInterface(pIBitmapDecoder,
                                          IID_IBitmapFrameWithSoftwareBitmap,
                                          pIBitmapFrameWithSoftwareBitmap) Then
                                If BitmapFrameWithSoftwareBitmap_GetSoftwareBitmapAsync(pIBitmapFrameWithSoftwareBitmap,
                                                                                    pISoftwareBitmap) Then
                                    bolRet = WaitForAsync(pISoftwareBitmap)
                                End If
                                Release(pIBitmapFrameWithSoftwareBitmap)
                            End If
                            If Not bolIsBitmapPixelFormatSupported Then
                                bolRet = False
                                Dim pISoftwareBitmapTmp As IntPtr = IntPtr.Zero
                                If SoftwareBitmapStatics_Convert(m_pISoftwareBitmapStatics,
                                                             pISoftwareBitmap,
                                                             pixelFormat,
                                                             pISoftwareBitmapTmp) Then
                                    CloseAndRelease(pISoftwareBitmap)
                                    pISoftwareBitmap = pISoftwareBitmapTmp
                                    bolRet = True
                                End If
                            End If
                        End If
                        Release(pIBitmapFrame)
                    End If
                End If
                Release(pIBitmapDecoder)
            End If
        End If
        Return bolRet
    End Function

    Private Function GetSupportedBitmapPixelFormats(pIFaceDetectorStatics As IntPtr) As Boolean
        Dim bolRet As Boolean = False
        Dim pIReadOnlyList As IntPtr = IntPtr.Zero
        If FaceDetectorStatics_GetSupportedBitmapPixelFormats(pIFaceDetectorStatics,
                                                              pIReadOnlyList) Then
            Dim intCount As Integer
            If ReadOnlyList_GetSize(pIReadOnlyList,
                                    intCount) Then
                If intCount > 0 Then
                    m_eBitmapPixelFormats = New List(Of BitmapPixelFormat)
                    For intIndex As Integer = 0 To intCount - 1
                        Dim eBitmapPixelFormat As BitmapPixelFormat
                        If ReadOnlyList_GetAt(pIReadOnlyList, intIndex, eBitmapPixelFormat) Then
                            m_eBitmapPixelFormats.Add(eBitmapPixelFormat)
                        End If
                    Next
                    bolRet = True
                End If
            End If
            Release(pIReadOnlyList)
        End If
        Return bolRet
    End Function

    Private Function GetActivationFactory(ClassName As String,
                                          IID As String,
                                          ByRef pFactory As IntPtr) As Boolean
        Dim bolRet As Boolean = False
        Dim hString As IntPtr = IntPtr.Zero
        If WindowsCreateString(ClassName, ClassName.Length, hString) = S_OK Then
            If hString <> IntPtr.Zero Then
                If RoGetActivationFactory(hString, New Guid(IID), pFactory) = S_OK Then bolRet = True
                WindowsDeleteString(hString)
            End If
        End If
        Return bolRet
    End Function

    Private Function WaitForAsync(ByRef pInterface As IntPtr) As Boolean
        Dim bolret As Boolean = False
        If pInterface <> IntPtr.Zero Then
            Dim pIAsyncInfo As IntPtr = IntPtr.Zero
            If QueryInterface(pInterface, IID_IAsyncInfo, pIAsyncInfo) Then
                Dim eStatus As AsyncStatus
                Do
                    AsyncInfo_GetStatus(pIAsyncInfo, eStatus)
                Loop While eStatus = AsyncStatus.Started
                If eStatus = AsyncStatus.Completed Then
                    bolret = True
                Else
                    AsyncInfo_Close(pIAsyncInfo)
                End If
                Release(pIAsyncInfo)
                If bolret Then
                    Dim pRet As IntPtr = IntPtr.Zero
                    If AsyncInterface_GetResults(pInterface, pRet) Then
                        Release(pInterface)
                        pInterface = pRet
                    End If
                End If
            End If
        End If
        Return bolret
    End Function

    Private Function GetVTablePtr(pInterface As IntPtr,
                                  intMethodNumber As VTable_Interfaces,
                                  ByRef pVTablePtr As IntPtr) As Boolean
        Dim bolRet As Boolean = False
        If pInterface <> IntPtr.Zero AndAlso intMethodNumber >= 0 Then
            Dim pVTablePtrs As IntPtr() = New IntPtr(0) {}
            Marshal.Copy(IntPtr.Add(Marshal.ReadIntPtr(pInterface),
                                    IntPtr.Size * intMethodNumber),
                                    pVTablePtrs, 0, 1)
            pVTablePtr = pVTablePtrs(0)
            bolRet = True
        End If
        Return bolRet
    End Function

    Private Sub Release(ByRef pInterface As IntPtr)
        If pInterface <> IntPtr.Zero Then
            Marshal.Release(pInterface)
            pInterface = IntPtr.Zero
        End If
    End Sub

    Private Sub CloseAndRelease(ByRef pInterface As IntPtr)
        If pInterface <> IntPtr.Zero Then
            Dim pIClosable As IntPtr = IntPtr.Zero
            If QueryInterface(pInterface, IID_IClosable, pIClosable) Then
                If Closable_Close(pIClosable) Then
                    Marshal.Release(pIClosable)
                    Marshal.Release(pInterface)
                    pInterface = IntPtr.Zero
                End If
            End If
        End If
    End Sub

    Private Function QueryInterface(pInterface As IntPtr,
                                    IID As String,
                                    ByRef pIOut As IntPtr) As Boolean
        Dim bolRet As Boolean = False
        If pInterface <> IntPtr.Zero Then
            If Marshal.QueryInterface(pInterface, New Guid(IID), pIOut) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function
#End Region

#Region "Interface Delegates"

    Private Delegate Function IReadOnlyList_GetAt(<[In]> this As IntPtr,
                                                  <[In]> index As Integer,
                                                  <Out> ByRef value As IntPtr) As Integer

    Private Delegate Function IReadOnlyList_GetAt_Enum(<[In]> this As IntPtr,
                                                       <[In]> index As Integer,
                                                       <Out> ByRef value As BitmapPixelFormat) As Integer

    Private Delegate Function IReadOnlyList_GetSize(<[In]> this As IntPtr,
                                                    <Out> ByRef value As Integer) As Integer


    Private Delegate Function IBitmapDecoderStatics_CreateAsync(<[In]> this As IntPtr,
                                                                <[In]> stream As IntPtr,
                                                                <Out> ByRef asyncInfo As IntPtr) As Integer

    Private Delegate Function IBitmapFrame_GetBitmapPixelFormat(<[In]> this As IntPtr,
                                                                <Out> ByRef value As BitmapPixelFormat) As Integer

    Private Delegate Function IBitmapFrameWithSoftwareBitmap_GetSoftwareBitmapAsync(<[In]> this As IntPtr,
                                                                                    <Out> ByRef value As IntPtr) As Integer

    Private Delegate Function ISoftwareBitmapStatics_Convert(<[In]> this As IntPtr,
                                                             <[In]> source As IntPtr,
                                                             <[In]> format As BitmapPixelFormat,
                                                             <Out> ByRef value As IntPtr) As Integer

    Private Delegate Function IFaceDetectorStatics_CreateAsync(<[In]> this As IntPtr,
                                                              <Out> ByRef asyncInfo As IntPtr) As Integer

    Private Delegate Function IFaceDetectorStatics_GetSupportedBitmapPixelFormats(<[In]> this As IntPtr,
                                                                                  <Out> ByRef result As IntPtr) As Integer

    Private Delegate Function IFaceDetector_DetectFacesAsync(<[In]> this As IntPtr,
                                                             <[In]> image As IntPtr,
                                                             <Out> ByRef returnValue As IntPtr) As Integer

    Private Delegate Function IDetectedFace_GetFaceBox(<[In]> this As IntPtr,
                                                       <Out> ByRef returnValue As Rectangle) As Integer

    Private Delegate Function IAsyncInfo_GetStatus(<[In]> this As IntPtr,
                                                    <Out> ByRef status As AsyncStatus) As Integer

    Private Delegate Function IAsyncInfo_Close(<[In]> this As IntPtr) As Integer

    Private Delegate Function IAsyncInterface_GetResults(<[In]> this As IntPtr,
                                                         <Out> ByRef results As IntPtr) As Integer

    Private Delegate Function IClosable_Close(<[In]> this As IntPtr) As Integer
#End Region

#Region "Interface Functions"
    Private Function ReadOnlyList_GetAt(this As IntPtr,
                                        index As Integer,
                                        ByRef value As IntPtr) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IReadOnlyList_GetAt, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IReadOnlyList_GetAt)),
                                                           IReadOnlyList_GetAt).Invoke(this, index, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function ReadOnlyList_GetAt(this As IntPtr,
                                        index As Integer,
                                        ByRef value As BitmapPixelFormat) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IReadOnlyList_GetAt, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IReadOnlyList_GetAt_Enum)),
                                                           IReadOnlyList_GetAt_Enum).Invoke(this, index, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function ReadOnlyList_GetSize(this As IntPtr,
                                          ByRef value As Integer) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IReadOnlyList_GetSize, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IReadOnlyList_GetSize)),
                                                           IReadOnlyList_GetSize).Invoke(this, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function BitmapDecoderStatics_CreateAsync(this As IntPtr,
                                                      stream As IntPtr,
                                                      ByRef asyncInfo As IntPtr) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IBitmapDecoderStatics_CreateAsync, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IBitmapDecoderStatics_CreateAsync)),
                                                           IBitmapDecoderStatics_CreateAsync).Invoke(this, stream, asyncInfo) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function BitmapFrame_GetBitmapPixelFormat(this As IntPtr,
                                                      ByRef value As BitmapPixelFormat) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IBitmapFrame_GetBitmapPixelFormat, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IBitmapFrame_GetBitmapPixelFormat)),
                                                           IBitmapFrame_GetBitmapPixelFormat).Invoke(this, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function BitmapFrameWithSoftwareBitmap_GetSoftwareBitmapAsync(this As IntPtr,
                                                                          ByRef value As IntPtr) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IBitmapFrameWithSoftwareBitmap_GetSoftwareBitmapAsync, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IBitmapFrameWithSoftwareBitmap_GetSoftwareBitmapAsync)),
                                                           IBitmapFrameWithSoftwareBitmap_GetSoftwareBitmapAsync).Invoke(this, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function SoftwareBitmapStatics_Convert(this As IntPtr,
                                                   source As IntPtr,
                                                   format As BitmapPixelFormat,
                                                   ByRef value As IntPtr) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.ISoftwareBitmapStatics_Convert, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(ISoftwareBitmapStatics_Convert)),
                                                           ISoftwareBitmapStatics_Convert).Invoke(this, source, format, value) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function FaceDetectorStatics_CreateAsync(this As IntPtr,
                                                     ByRef asyncInfo As IntPtr) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IFaceDetectorStatics_CreateAsync, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IFaceDetectorStatics_CreateAsync)),
                                                           IFaceDetectorStatics_CreateAsync).Invoke(this, asyncInfo) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function FaceDetectorStatics_GetSupportedBitmapPixelFormats(this As IntPtr,
                                                                        ByRef result As IntPtr) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IFaceDetectorStatics_GetSupportedBitmapPixelFormats, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IFaceDetectorStatics_GetSupportedBitmapPixelFormats)),
                                                           IFaceDetectorStatics_GetSupportedBitmapPixelFormats).Invoke(this, result) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function FaceDetector_DetectFacesAsync(this As IntPtr,
                                                   image As IntPtr,
                                                   ByRef returnValue As IntPtr) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IFaceDetector_DetectFacesAsync, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IFaceDetector_DetectFacesAsync)),
                                                           IFaceDetector_DetectFacesAsync).Invoke(this, image, returnValue) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function DetectedFace_GetFaceBox(this As IntPtr,
                                             ByRef returnValue As Rectangle) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IDetectedFace_GetFaceBox, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IDetectedFace_GetFaceBox)),
                                                           IDetectedFace_GetFaceBox).Invoke(this, returnValue) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function AsyncInfo_GetStatus(this As IntPtr,
                                         ByRef status As AsyncStatus) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IAsyncInfo_GetStatus, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IAsyncInfo_GetStatus)),
                                                           IAsyncInfo_GetStatus).Invoke(this, status) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function AsyncInfo_Close(this As IntPtr) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IAsyncInfo_Close, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IAsyncInfo_Close)),
                                                           IAsyncInfo_Close).Invoke(this) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function AsyncInterface_GetResults(this As IntPtr,
                                               ByRef results As IntPtr) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IAsyncInterface_GetResults, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IAsyncInterface_GetResults)),
                                                           IAsyncInterface_GetResults).Invoke(this, results) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function

    Private Function Closable_Close(this As IntPtr) As Boolean
        Dim pVTablePtr As IntPtr
        Dim bolRet As Boolean = False
        If this <> IntPtr.Zero AndAlso GetVTablePtr(this, VTable_Interfaces.IClosable_Close, pVTablePtr) Then
            If CType(Marshal.GetDelegateForFunctionPointer(pVTablePtr, GetType(IClosable_Close)),
                                                           IClosable_Close).Invoke(this) = S_OK Then bolRet = True
        End If
        Return bolRet
    End Function
#End Region

#Region "Implements IDisposable"
    Protected Overridable Sub Dispose(disposing As Boolean)
        If Not m_bolDisposedValue Then
            If disposing Then
            End If
            Release(m_pIFaceDetector)
            Release(m_pISoftwareBitmapStatics)
            Release(m_pIBitmapDecoderStatics)
            m_bolDisposedValue = True
        End If
    End Sub

    Public Sub Dispose() Implements IDisposable.Dispose
        Dispose(disposing:=True)
        GC.SuppressFinalize(Me)
    End Sub
#End Region

End Class
